/*
 * mrp - a program for multiple regression (mrp 1.9)
 *
 * Copyright (c) Heinz Spiess, CH-2558 Aegerten, Switzerland, 1988.
 * All rights reserved.
 *
 */
static char    *sccs = "@(#)mrp	1.9";

#include <stdio.h>
#include <math.h>
#define inline InLine
#define MAXLINE 1024     /* maximal length of input line */
#define MAXDIM 25        /* maximal number of variables, incl 1s and Y */
#define MAXITEM 128      /* maximal number of items per line in input file */
int             n=0;	/* number of elements */
int             inline=0;	/* number of input lines */
int		echo=0;		/* echo input values */
int		debug=0;	/* display debug information */
int		silent=0;	/* ignore uncoplete input lines silently */
int		printxx=0;	/* print X'X matrix */
int		quiet=0;        /* short form of output */
int 		noconst=0;      /* no constant term */

main(argc, argv)
    char          **argv;
    int             argc;
{
    FILE           *input;
    char           *c;
    static char	    name[MAXDIM][9];
    static double   xx[MAXDIM][MAXDIM];
    static double   xxi[MAXDIM][2*MAXDIM];
    static double   x[MAXDIM];
    static double   sumx[MAXDIM];
    static double   y, yy, sst, ssr, sse, r2, s2, sigma, t, vi;
    double          sqrt();
    int             index[MAXDIM], nv, nmax, l, k, maxind = 0;
    int		    inthand();
    register int    i,j;

    input = stdin;
    setbuf(input, NULL);
    nmax = 0;
    index[0] = 0;
    strcpy(name[0],"CONSTANT");

    while (argc>1 && argv[1][0] == '-') {
	c = argv[1] + 1;
	argv++;
	argc--;
	while (*c)
	    switch (*(c++)) {
	    case '0':
		nmax = -1;
		noconst = 1;
		break;
	    case 'x':
		printxx = 1;
		break;
	    case 'e':
		echo = 1;
		break;
	    case 'h':
		argc=0;
		break;
	    case 's':
		silent = 1;
		break;
    	    case 'q':
		quiet = 1;
		break;
    	    case 'i':
		signal(2,inthand);
		break;
	    case 'd':
		debug = 1;
	    }
    }
    if(!quiet) printf(
        "\n%s - (c) Heinz Spiess, CH-2558 Aegerten, 1988-92.  All rights reserved.\n\n"
       ,"mrp 1.9"); 
    if(argc<=2)usage(0);
    while (--argc) {
	nmax++;
	index[nmax]=999999;
	name[nmax][0]='\0';
	sscanf(argv[1],"%d=%8s",&index[nmax],name[nmax]);
	if(index[nmax]<1 || index[nmax]>MAXDIM)usage(1);
	if(!name[nmax][0])sprintf(name[nmax],"VAR%d",index[nmax]);
	if (index[nmax] > maxind)
	    maxind = index[nmax];
	argv++;
    }

    for (i = 0; i <= nmax; i++){
	sumx[i]=0;
	for (j = 0; j <= nmax; j++)
	    xx[i][j] = 0.;
	}
    l = 0;
    y = 0.;
    yy = 0.;
    while (nv = getvec(x, input, index, nmax)) {
	l++;
	if (nv < nmax) {
	    if(!silent)
	       fprintf(stderr, "input error on line %d (field %d)\n", l, nv);
	    continue;
	}
	n++;
	y += x[nmax];
	yy += x[nmax] * x[nmax];
	if (echo)
	    printf("%4d: ", l);
	for (i = 0; i <= nmax; i++) {
	    register double xv= x[i];
	    sumx[i]+=xv;
	    if (echo)
		printf("%9.4f ", xv);
	    for (j = i; j <= nmax; j++)
		xx[i][j] += xv * x[j];
	}
	if (echo)
	    printf("\n");
    }

    if(!quiet)printf("\nInput lines: %d read, %d accepted, %d skipped.\n",
	inline,n,inline-n);

    /* symmetrize xx */
    for(i=0;i<=nmax;i++)
	for(j=i+1;j<=nmax;j++)
	    xx[j][i]=xx[i][j];

    if (printxx){
	printf("\nMATRIX X'(X|Y):");
	printmat(xx, nmax , nmax + 1, name);
    }

    if(!quiet){
      printf("\nCORRELATION:  ");
      {
    	int ij0=noconst?0:1;
        for (j = ij0; j <= nmax; j++)printf(" %8.8s",name[j]);
        for (i = ij0; i <= nmax; i++) {
		printf("\n%-8.8s      ",name[i]);
        	for (j = ij0; j<=i; j++){
		    double rij,denom;
		    rij=(j==i)?1:0;
		    denom=sqrt((xx[i][i]-sumx[i]*sumx[i]/n)*
			     (xx[j][j]-sumx[j]*sumx[j]/n));
		    if(denom>0)rij=(xx[i][j]-sumx[i]*sumx[j]/n)/denom;
		    printf(" %8g",rij);
		}
	}
      }
    }

    for (i = 0; i < nmax; i++){
	for (j = 0; j <= nmax; j++)
	    xxi[i][j] = xx[i][j];
	for (j = nmax+1; j <= 2*nmax; j++)
	    xxi[i][j] = (j==i+nmax+1 ? 1 : 0);
	}

    lineq(xxi, nmax, 2*nmax + 1);

    /* calculate R2 */
    sst = yy;
    ssr = 0;
    if (!noconst) {
        sst -= y * y / n;
        ssr -= y * y / n;
    }
    for (i = 0; i < nmax; i++)
	ssr += xxi[i][nmax] * xx[i][nmax];
    sse = sst - ssr;
    r2 = ssr / sst;
    s2 = sse / (n - nmax);
    sigma = sqrt(s2);
    if(!quiet){

        printf("\n\nREGRESSION COEFFICIENTS FOR %s:\n",name[nmax]);
        for (i = 0; i < nmax; i++) {
	    t=1e37;
	    vi=s2*xxi[i][nmax+i+1];
	    if(vi>0)t=xxi[i][nmax]/sqrt(vi);
	    printf("B%02d(%8.8s)=%18f (t<%d>=%g)\n",
		i, name[i], xxi[i][nmax],n-nmax,t);
        }

	printf("\nCOVARIANCE:   ");
        for (j = 0; j<nmax; j++)printf(" B%02d(%8.8s)",j,name[j]);
        for (i = 0; i < nmax; i++) {
		printf("\nB%02d(%8.8s) ",i,name[i]);
        	for (j = 0; j<=i; j++)
		    printf(" %13g",s2*xxi[i][nmax+j+1]);
	}

        printf("\n\nANALYSIS OF VARIANCE FOR %s:\n",name[nmax]);
        printf("SST=%18f      DF=%4d\n", sst, n - 1);
        printf("SSR=%18f      DF=%4d\n", ssr, nmax - 1);
        printf("SSE=%18f      DF=%4d\n", sse, n - nmax );
        printf("R^2=%18f\n", r2);
        printf("SIG^2=%16f\n", s2);
        printf("SIGMA=%16f\n", sigma);
    } else {
	printf("%5d %8.5f %12f ",n,r2,sigma);
        for (i = 0; i < nmax; i++) 
	    printf("%12f ",xxi[i][nmax]);
	printf("\n");
    }
}

getvec(x, input, index, nmax)
double          x[];
int		index[];
int		nmax;
FILE           *input;
{
    int             i, j, nit;
    char            line[MAXLINE];
    char           *item[MAXITEM];
    x[0] = 1.;
    if (fgets(line, MAXLINE, input) == NULL)
	return (0);
    inline++;
    for(nit=0,j=0;line[j];j++){
	if(line[j]==' '||line[j]=='\t')line[j]='\0';
	else if(j==0||!line[j-1])item[nit++]=line+j;
    }
    if(debug)fprintf(stderr,"Line %d - %d items\n",inline,nit);
    for(i=index[0]?0:1;i<=nmax;i++){
       if(index[i]>nit||sscanf(item[index[i]-1],"%lf",&x[i])!=1)
	   return(-index[i]);
    }
    return (nmax);
}

lineq(xx, m, n)
    int             m, n;
    double          xx[MAXDIM][2*MAXDIM];
{
    int             i, j, k;
    for (i = 0; i < m; i++) {
	if (xx[i][i] == 0){
	    fprintf(stderr, "Error in line %d - X'X not full rank!\n", i);
	} else {
	    for (j = i + 1; j < n; j++)
	        xx[i][j] /= xx[i][i];
	    xx[i][i] = 1.;
	    for (k = 0; k < m; k++) {
	        if (k == i)
		    continue;
	        for (j = i + 1; j < n; j++)
		    xx[k][j] -= xx[k][i] * xx[i][j];
	        xx[k][i] = 0;
	    }
	}
	if (debug)
	    printmat(xx, m, n, NULL);
    }
}

printmat(xx, m, n, name)
    int             m, n;
    double          xx[MAXDIM][MAXDIM];
    char 	    name[MAXDIM][9];
{
    int             i, j;
    printf("\n         ");
    if(name) for (j = 0; j < n; j++) printf("     %8s",name[j]);
    for (i = 0; i < m; i++) {
	if(name)printf("\n%8.8s  ",name[i]);
	for (j = 0; j < n; j++)
	    printf("%12.4f ", xx[i][j]);
    }
    printf("\n");
}

inthand(ic)
int ic;
{
	fprintf(stderr,"\nmrpr %d input lines read, %d skipped.\n",
		inline,inline-n);
	signal(ic,inthand);
}

usage(key)
int key;
{
char *text="\n\
MRP is a program for performing multiple linear regression analysis.\n\
\n\
Usage:   mrp [options] [field numbers] <[input file]\n\
\n\
Each line of standard input contains a set of values. The command line\n\
arguments are the numbers of the input fields that will be used for the\n\
multiple regression, the last field is taken to be the dependent variable. \n\
E.g. the command `mrp 1 3 5 2 <input' will compute a multiple regression\n\
to express field number 2 (the last argument) as a linear combination of\n\
fields 1, 3 and 5 (plus a constant term, unless option -0 is used). \n\
\n\
The field numbers for dependent and independent variables may optionally\n\
be followed by a clause '=<name>', in which case this name is used to\n\
annotate the output.  Example: `mrp 3=VOLAM 4=VOLPM 15=VOLOBS <input'\n\
(The names VOLAM and VOLPM are used for the independent variables in fields\n\
3 and 4, while the dependent variable in field 15 is named VOLOBS.)\n\
\n\
The following command line options are available:\n\
   -0   Perform regression without constant term. \n\
   -d   Print out debug information. \n\
   -e   Echo input data when it is read in. \n\
   -h   Displays this message.\n\
   -i   Catch interrupt (signal 2) and display status information.\n\
   -s   Ignore incomplete input lines silently.\n\
   -x   Print matrix X'X.\n\
   -q   Quiet - produce only one line of output that contains the\n\
        number of data sets accepted, the value of R^2, sigma and the\n\
        regression coefficients, suitable for further processing by awk.\n\
        \n\
 Currently, the following limitations apply:\n\
   Maximum number of fields in input line:      128\n\
   Maximum number of independent variables:      24\n\
   Maximum length of variable name:               8\n";
printf(text);
exit(key);
}
